#include "cmdline.h"
#include "connection.h"
#include "jbuffer.h"
#include "request.h"
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>

static void *worker ();
static int cloexec_flag (int fd);

static BNDBUF *bndbufbuf;

int
initConnectionHandler (void)
{
  const char *threads_char = cmdlineGetValueForKey ("threads");
  int threads = 6;
  if (threads_char != NULL)
    {
      errno = 0;
      char *endstr;
      threads = strtol (threads_char, &endstr, 10);
      if (errno != 0)
        {
          perror ("strtol");
          return -1;
        }
      if (endstr[0] != '\0')
        {
          fprintf (stderr, "invalid threads number\n");
          return -1;
        }
    }

  const char *bufsize_char = cmdlineGetValueForKey ("bufsize");
  size_t bufsize = 8;
  if (bufsize_char != NULL)
    {
      errno = 0;
      char *endstr;
      bufsize = strtol (bufsize_char, &endstr, 10);
      if (errno != 0)
        {
          perror ("strtol");
          return -1;
        }
      if (endstr[0] != '\0')
        {
          fprintf (stderr, "invalid buffer size\n");
          return -1;
        }
    }

  bndbufbuf = bbCreate (bufsize);
  if (bndbufbuf == NULL)
    {
      fprintf (stderr, "bbCreate failed with size %i", (int)bufsize);
      return -1;
    }

  struct sigaction sig_action = {
    .sa_flags = SA_RESTART,
    .sa_handler = SIG_IGN,
  };
  if (sigemptyset (&sig_action.sa_mask) == -1)
    {
      perror ("sigemptyset");
    }
  if (sigaction (SIGPIPE, &sig_action, NULL) == -1)
    {
      perror ("sigaction");
      return -1;
    }

  if (initRequestHandler () == -1)
    {
      return -1;
    }

  for (int i = 0; i < threads; i++)
    {
      pthread_t thread;
      int tmp_errno = pthread_create (&thread, NULL, &worker, NULL);
      if (tmp_errno != 0)
        {
          errno = tmp_errno;
          perror ("pthread_create");
          return -1;
        }
    }

  return 0;
}

void
handleConnection (int clientSock, int listenSock)
{
  if (cloexec_flag (clientSock) == -1)
    {
      return;
    }
  if (cloexec_flag (listenSock) == -1)
    {
      return;
    }

  bbPut (bndbufbuf, clientSock);
}

static void *
worker ()
{
  int errno_tmp = pthread_detach (pthread_self ());
  if (errno_tmp != 0)
    {
      errno = errno_tmp;
      perror ("pthread_detach");
    }

  while (1)
    {
      int clientSock = bbGet (bndbufbuf);

      FILE *rx = fdopen (clientSock, "r");
      if (rx == NULL)
        {
          perror ("fdopen");
          close (clientSock);
          continue;
        }

      int clientSockWrite = dup (clientSock);
      if (clientSockWrite == -1)
        {
          perror ("dup");
          fclose (rx);
          continue;
        }

      if (cloexec_flag (clientSockWrite) == -1)
        {
          fclose (rx);
          close (clientSockWrite);
          continue;
        }

      FILE *tx = fdopen (clientSockWrite, "w");
      if (tx == NULL)
        {
          perror ("fdopen");
          fclose (rx);
          close (clientSockWrite);
          continue;
        }

      handleRequest (rx, tx);

      if (fclose (rx) == EOF)
        {
          perror ("fclose");
        }
      if (fclose (tx) == EOF)
        {
          perror ("fclose");
        }
    }
  return NULL;
}

static int
cloexec_flag (int fd)
{
  int flags = fcntl (fd, F_GETFD, 0);
  if (flags == -1)
    {
      perror ("fcntl");
      return -1;
    }

  if (fcntl (fd, F_SETFD, flags | FD_CLOEXEC) == -1)
    {
      perror ("fcntl");
      return -1;
    }

  return 0;
}
