#include <errno.h>
#include <netdb.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

static void getresponse (FILE *response, char *status);
static void makecall (FILE *call, int ret_code);

static void
die (char *msg)
{
  perror (msg);
  exit (EXIT_FAILURE);
}

int
main (int argc, char *argv[])
{
  if (argc < 2)
    {
      fprintf (stderr, "Wrong Input: snail [-s <subject>] <address>\n");
      exit (EXIT_FAILURE);
    }

  int subject = !strcmp (argv[1], "-s");

  if ((subject && argc != 4) || (!subject && argc != 2))
    {
      fprintf (stderr, "Wrong Input: snail [-s <subject>] <address>\n");
      exit (EXIT_FAILURE);
    }

  char *subject_char = "";
  if (subject)
    {
      subject_char = argv[2];
    }

  char *address_char = "";
  if (subject)
    {
      address_char = argv[3];
    }
  else
    {
      address_char = argv[1];
    }

  char *body = malloc (1);
  if (body == NULL)
    {
      die ("malloc");
    }

  int c;
  int index = 0;
  while (!feof (stdin))
    {
      c = fgetc (stdin);
      if (c == EOF)
        {
          break;
        }

      if (c == '\n' && ((index > 0 && body[index - 1] != '\r') || index == 0))
        {
          body = realloc (body, index + 2);
          if (body == NULL)
            {
              die ("realloc");
            }
          body[index++] = '\r';
        }
      else if (c == '.'
               && ((index > 0 && body[index - 1] == '\n') || index == 0))
        {
          body = realloc (body, index + 2);
          if (body == NULL)
            {
              die ("realloc");
            }
          body[index++] = '.';
        }
      else
        {
          body = realloc (body, index + 1);
          if (body == NULL)
            {
              die ("realloc");
            }
        }

      body[index++] = c;
    }

  if (ferror (stdin))
    {
      die ("fgetc");
    }

  body = realloc (body, index + 6);
  if (body == NULL)
    {
      die ("realloc");
    }
  body[index++] = '\r';
  body[index++] = '\n';
  body[index++] = '.';
  body[index++] = '\r';
  body[index++] = '\n';
  body[index++] = '\0';

  uid_t uid = getuid ();
  errno = 0;
  struct passwd *pwuid = getpwuid (uid);
  if (pwuid == NULL)
    {
      if (errno != 0)
        {
          die ("getpwuid");
        }
      else
        {
          fprintf (stderr, "getpwuid failed\n");
          exit (EXIT_FAILURE);
        }
    }
  char *username_tmp = pwuid->pw_name;
  char username[strlen (username_tmp) + 1];
  strncpy (username, username_tmp, sizeof (username));

  char *fullname_tmp = pwuid->pw_gecos;
  char fullname2[strlen (fullname_tmp) + 1];
  strncpy (fullname2, fullname_tmp, sizeof (fullname2));
  char *fullname = strtok (fullname2, ",");
  if (fullname == NULL)
    {
      fprintf (stderr, "pw_gecos is malformed\n");
      exit (EXIT_FAILURE);
    }

  char hostname[256];
  if (gethostname (hostname, sizeof (hostname)) == -1)
    {
      die ("gethostname");
    }

  struct addrinfo fqdn_hints = {
    .ai_flags = AI_CANONNAME,
  };

  struct addrinfo *fqdn_head;
  int ret_addrinfo = getaddrinfo (hostname, NULL, &fqdn_hints, &fqdn_head);
  if (ret_addrinfo == EAI_SYSTEM)
    {
      die ("getaddrinfo");
    }
  else if (ret_addrinfo != 0)
    {
      fprintf (stderr, "getaddrinfo error: %s\n", gai_strerror (ret_addrinfo));
      exit (EXIT_FAILURE);
    }

  char *fqdn = fqdn_head->ai_canonname;

  struct addrinfo hints = {
    .ai_socktype = SOCK_STREAM,
    .ai_family = AF_UNSPEC,
    .ai_flags = AI_ADDRCONFIG,
  };
  struct addrinfo *head;

  ret_addrinfo = getaddrinfo ("faui03.cs.fau.de", "25", &hints, &head);
  if (ret_addrinfo == EAI_SYSTEM)
    {
      die ("getaddrinfo");
    }
  else if (ret_addrinfo != 0)
    {
      fprintf (stderr, "getaddrinfo error: %s\n", gai_strerror (ret_addrinfo));
      exit (EXIT_FAILURE);
    }

  int sock;
  struct addrinfo *current;
  for (current = head; current != NULL; current = current->ai_next)
    {
      sock = socket (current->ai_family, current->ai_socktype,
                     current->ai_protocol);
      if (sock == -1)
        {
          perror ("socket");
          continue;
        }

      int ret_con = connect (sock, current->ai_addr, current->ai_addrlen);
      if (ret_con == -1)
        {
          perror ("connect");
        }
      else if (ret_con == 0)
        {
          break;
        }
      else
        {
          if (close (sock) == -1)
            {
              perror ("close");
            }
        }
    }
  if (current == NULL)
    {
      fprintf (stderr, "keine brauchbare Adresse gefunden\n");
      exit (EXIT_FAILURE);
    }

  FILE *call = fdopen (sock, "w");
  if (call == NULL)
    {
      die ("fdopen");
    }
  int sock_copy = dup (sock);
  if (sock_copy == -1)
    {
      die ("dup");
    }
  FILE *response = fdopen (sock_copy, "r");
  if (response == NULL)
    {
      die ("fdopen");
    }

  getresponse (response, "220 ");

  makecall (call, fprintf (call, "HELO %s\r\n", fqdn));

  getresponse (response, "250 ");

  makecall (call,
            fprintf (call, "MAIL FROM:<%s@cip.cs.fau.de>\r\n", username));

  getresponse (response, "250 ");

  makecall (call, fprintf (call, "RCPT TO:<%s>\r\n", address_char));

  getresponse (response, "250 ");

  makecall (call, fprintf (call, "DATA\r\n"));

  getresponse (response, "354 ");

  makecall (call, fprintf (call, "From: %s <%s@cip.cs.fau.de>\r\nTo: <%s>\r\n",
                           fullname, username, address_char)
                      < 0);

  if (subject)
    {
      makecall (call, fprintf (call, "Subject: %s\r\n", subject_char));
    }

  makecall (call, fprintf (call, "\r\n%s", body));

  getresponse (response, "250 ");

  makecall (call, fprintf (call, "QUIT\r\n"));

  getresponse (response, "221 ");

  if (fclose (call) == EOF)
    {
      die ("fclose");
    }
  if (fclose (response) == EOF)
    {
      die ("fclose");
    }
  freeaddrinfo (head);
  freeaddrinfo (fqdn_head);
  free (body);
  return EXIT_SUCCESS;
}

static void
makecall (FILE *call, int ret_code)
{
  if (ret_code < 0)
    {
      fprintf (stderr, "error in fprintf");
      exit (EXIT_FAILURE);
    }
  if (fflush (call) == EOF)
    {
      die ("fflush");
    }
}

static void
getresponse (FILE *response, char *status)
{
  char *message = malloc (1);
  if (message == NULL)
    {
      die ("malloc");
    }

  int c;
  int index = 0;
  while (!feof (response))
    {
      c = fgetc (response);
      if (c == EOF || c == '\n')
        {
          break;
        }

      message = realloc (message, index + 1);
      if (message == NULL)
        {
          die ("realloc");
        }

      message[index++] = c;
    }

  if (ferror (response))
    {
      die ("fgetc");
    }

  message = realloc (message, index + 1);
  if (message == NULL)
    {
      die ("realloc");
    }
  message[index] = '\0';

  if (index <= 3)
    {
      fprintf (stderr,
               "recieved response without three digit status code: %s\n",
               message);
      exit (EXIT_FAILURE);
    }

  if (strncmp (status, message, 4) != 0)
    {
      fprintf (stderr, "recieved wrong status code (expected %s): %s\n",
               status, message);
      exit (EXIT_FAILURE);
    }

  free (message);
}
