#define _GNU_SOURCE
#include "plist.h"
#include "shellutils.h"
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

static volatile int foreground_end;

static int
printProcEvent (const ProcInfo *info, int event)
{
  pid_t pid = info->pid;
  const char *cmdLine = info->cmdLine;

  fprintf (stderr, "[%d] \"%s\" ", pid, cmdLine);
  if (WIFEXITED (event) != 0)
    fprintf (stderr, "exited with status %d", WEXITSTATUS (event));
  else if (WIFSIGNALED (event) != 0)
    fprintf (stderr, "terminated by signal %d", WTERMSIG (event));
  else if (WIFSTOPPED (event) != 0)
    fprintf (stderr, "stopped by signal %d", WSTOPSIG (event));
  else if (WIFCONTINUED (event) != 0)
    fputs ("continued", stderr);
  putc ('\n', stderr);

  if (plistRemove (pid) == -1)
    {
      fprintf (stderr,
               "No process with the given PID %i was found in the list.\n",
               (int)pid);
      return -1;
    }

  return 0;
}

static void
die (const char message[])
{
  perror (message);
  exit (EXIT_FAILURE);
}

static void
changeCwd (char *argv[])
{
  if (argv[1] == NULL || argv[2] != NULL)
    {
      fprintf (stderr, "Usage: %s <dir>\n", argv[0]);
      return;
    }
  if (chdir (argv[1]) != 0)
    perror (argv[0]);
}

static int
printProcInfo (const ProcInfo *info)
{
  fprintf (stderr, "[%d] \"%s\"\n", info->pid, info->cmdLine);
  return 0;
}

static void
printJobs (char *argv[])
{
  if (argv[1] != NULL)
    {
      fprintf (stderr, "Usage: %s\n", argv[0]);
      return;
    }
  plistIterate (printProcInfo);
}

static int
kill_callback (const ProcInfo *info)
{
  if (kill (info->pid, SIGTERM) == -1)
    {
      die ("kill");
    }
  return 0;
}

static void
signal_handler (int signal)
{
  int tmp_errno = errno;

  int status;
  pid_t child;
  while ((child = waitpid (-1, &status, WNOHANG)) != 0)
    {
      if (child == -1)
        {
          if (errno == ECHILD)
            {
              break;
            }
          die ("waitpid");
        }

      const ProcInfo *info = plistGet (child);
      if (info == NULL)
        {
          char *msg = "No process with the given PID was found in the list\n";
          if (write (STDERR_FILENO, msg, strlen (msg) + 1) == -1)
            {
              perror ("write");
            }
        }

      if (!info->background)
        {
          foreground_end = 1;
        }

      plistNotifyEvent (child, status);
    }

  errno = tmp_errno;
}

static void
execute (ShCommand *cmd)
{
  sigset_t oldMask;
  sigset_t mask;
  if (sigemptyset (&mask) == -1)
    {
      perror ("sigemptyset");
    }
  if (sigaddset (&mask, SIGCHLD) == -1)
    {
      die ("sigaddset");
    }
  if (sigprocmask (SIG_BLOCK, &mask, &oldMask) == -1)
    {
      die ("sigprocmask");
    }

  pid_t pid = fork ();
  if (pid == -1)
    die ("fork");

  if (pid == 0)
    {
      if (cmd->inFile != NULL)
        {
          int fd_in = open (cmd->inFile, O_RDONLY);
          if (fd_in == -1)
            {
              die ("open");
            }
          if (dup2 (fd_in, STDIN_FILENO) == -1)
            {
              die ("dup2");
            }
        }

      if (cmd->outFile != NULL)
        {
          int fd_out
              = creat (cmd->outFile, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
          if (fd_out == -1)
            {
              die ("creat");
            }
          if (dup2 (fd_out, STDOUT_FILENO) == -1)
            {
              die ("dup2");
            }
        }

      execvp (cmd->argv[0], cmd->argv);
      die (cmd->argv[0]);
    }
  else
    {
      const ProcInfo *info = plistAdd (pid, cmd->cmdLine, cmd->background);
      if (info == NULL)
        die ("plistAdd");

      if (cmd->background == true)
        {
          if (sigprocmask (SIG_SETMASK, &oldMask, NULL) == -1)
            {
              die ("sigprocmask");
            }
          printProcInfo (info);
        }
      else
        {
          foreground_end = 0;

          while (foreground_end == 0)
            {
              sigsuspend (&oldMask);
              if (errno != EINTR)
                {
                  die ("sigsuspend");
                }
            }
          if (sigprocmask (SIG_SETMASK, &oldMask, NULL) == -1)
            {
              die ("sigprocmask");
            }
        }
    }
}

int
main (void)
{

  struct sigaction sig_action = {
    .sa_handler = &signal_handler,
    .sa_flags = SA_RESTART,

  };
  if (sigfillset (&sig_action.sa_mask) == -1)
    {
      perror ("sigfillset");
    }
  if (sigaction (SIGCHLD, &sig_action, NULL) == -1)
    {
      die ("sigaction");
    }

  for (;;)
    {
      plistHandleEvents (&printProcEvent);

      shPrompt ();

      char cmdLine[MAX_CMDLINE_LEN];
      if (fgets (cmdLine, sizeof (cmdLine), stdin) == NULL)
        {
          if (ferror (stdin) != 0)
            die ("fgets");
          break;
        }

      ShCommand *cmd = shParseCmdLine (cmdLine);
      if (cmd == NULL)
        {
          perror ("shParseCmdLine");
          continue;
        }

      if (cmd->parseError != NULL)
        {
          if (strlen (cmdLine) > 0)
            fprintf (stderr, "Syntax error: %s\n", cmd->parseError);
        }
      else if (strcmp ("cd", cmd->argv[0]) == 0)
        {
          changeCwd (cmd->argv);
        }
      else if (strcmp ("jobs", cmd->argv[0]) == 0)
        {
          sigset_t oldMask;
          sigset_t mask;
          if (sigemptyset (&mask) == -1)
            {
              perror ("sigemptyset");
            }
          if (sigaddset (&mask, SIGCHLD) == -1)
            {
              die ("sigaddset");
            }
          if (sigprocmask (SIG_BLOCK, &mask, &oldMask) == -1)
            {
              die ("sigprocmask");
            }
          printJobs (cmd->argv);
          if (sigprocmask (SIG_SETMASK, &oldMask, NULL) == -1)
            {
              die ("sigprocmask");
            }
        }
      else if (strcmp ("nuke", cmd->argv[0]) == 0)
        {
          sigset_t oldMask;
          sigset_t mask;
          if (sigemptyset (&mask) == -1)
            {
              perror ("sigemptyset");
            }
          if (sigaddset (&mask, SIGCHLD) == -1)
            {
              die ("sigaddset");
            }
          if (sigprocmask (SIG_BLOCK, &mask, &oldMask) == -1)
            {
              die ("sigprocmask");
            }

          if (cmd->argv[1] == NULL)
            {
              plistIterate (&kill_callback);
            }
          else
            {
              errno = 0;

              char *endptr;
              long x = strtol (cmd->argv[1], &endptr, 10);
              if (errno != 0)
                {
                  die ("strtol");
                }
              if (cmd->argv[1] == endptr || *endptr != '\0')
                {
                  fprintf (stderr, "false number\n");
                  if (sigprocmask (SIG_SETMASK, &oldMask, NULL) == -1)
                    {
                      die ("sigprocmask");
                    }
                  continue;
                }

              const ProcInfo *info = plistGet ((pid_t)x);
              if (info == NULL)
                {
                  fprintf (stderr, "PID %li wurde nich gefunden\n", x);
                  if (sigprocmask (SIG_SETMASK, &oldMask, NULL) == -1)
                    {
                      die ("sigprocmask");
                    }
                  continue;
                }

              kill_callback (info);
            }

          sigset_t fullmask;
          if (sigfillset (&fullmask) == -1)
            {
              die ("sigfillset");
            }
          if (sigprocmask (SIG_BLOCK, &fullmask, NULL) == -1)
            {
              die ("sigprocmask");
            }
          sleep (1);

          if (sigprocmask (SIG_SETMASK, &oldMask, NULL) == -1)
            {
              die ("sigprocmask");
            }
        }
      else
        {
          execute (cmd);
        }
    }

  putchar ('\n');
  return EXIT_SUCCESS;
}